// Generated from Flocc program
// see http://www.flocc.net/

//#define DEBUG 1
//#define DEBUG 1
#include <mpi.h>
#include <vector>
#include <cstdlib>
#include <memory>
#include <time.h> // to seed srand
#include <cstring> // for strcmp
#include <cstdio> // sprintf
#include "arrays.h"
#include "cartTopologies.h"

#define N 3000

// BASE TYPES 

typedef struct nullStruct {
  bool operator<(const nullStruct& b) const {
    return false;
  }
  bool operator==(const nullStruct& b) const {
    return true;
  }
} nullStruct;

// PRINTING

void printVal(int v) { printf("%i", v); }
void printVal(const char *v) { printf("%s", v); }
void printVal(nullStruct v) {
  printVal("null");
}
void printVal(int* v) { 
  // only safe to print one, as don't know array length
  printVal(*v); 
} 

// HASHING

unsigned hashData ( const void *key, int len )
{
    unsigned char *p = (unsigned char*)key;
    unsigned h = 0;
    int i;
  
    for ( i = 0; i < len; i++ ) {
      h += p[i];
      h += ( h << 10 );
     h ^= ( h >> 6 );
   }
 
   h += ( h << 3 );
   h ^= ( h >> 11 );
   h += ( h << 15 );
 
   return h;
}

//#if __cplusplus <= 199711L
#ifndef __GXX_EXPERIMENTAL_CXX0X__
 // #error This library needs at least a C++11 compliant compiler

// if not C++ 11, define our own
// hash functions
namespace std {
   template<typename K>
   struct hash {
   };
}

#endif


// MPI STUFF

typedef struct NodeLayout {
  // all partition root nodes
  // in this node set
  MPI::Comm *partitionCom;
  int partitionCount, // number of partitions
      partitionRank;  // partition rank of this node

  // all nodes to mirror on in this partition
  MPI::Comm *mirrorCom;
  int mirrorCount, // number of nodes in mirror set
      mirrorRank,  // rank in mirror set of this node
      mirrorRootRank;    // rank of root in mirror set

  // if mirrorRank == mirrorRootRank, then this is a partition
  // head/root node

  // if partitionRank or mirrorRank is -1
  // then this node is not in this node layout

  bool operator==(const NodeLayout& b) const {
    // comparing layouts on a single machine, it should
    // be sufficient to compare communicators, as no other
    // layout should exist ON THIS MACHINE with the same
    // communicators.
    return (partitionCom == b.partitionCom) &&
           (mirrorCom == b.mirrorCom);
  }

  void Free() {
    // free communicators, 
    // ONLY SAFE AT VERY END, as other
    // layouts might use one of the same communicators.
    if (partitionCom != NULL && (*partitionCom) != MPI::COMM_WORLD) partitionCom->Free();
    if (mirrorCom != NULL) mirrorCom->Free();
  }

} NodeLayout;

NodeLayout allNodes;

#define NL_ON_FRINGE(nl) ((nl.mirrorRank == nl.mirrorRootRank) && (nl.mirrorRank >= 0))

// converts the rank of a process in one communicator to it's rank
// in another communicator.
void translateRank(MPI::Comm *srcComm, int srcRank, MPI::Comm *destComm, int *destRank) {
  MPI_Group orig_group, new_group; 
  orig_group = srcComm->Get_group();
  new_group = destComm->Get_group();
  MPI::Group::Translate_ranks(orig_group, 1, &srcRank, new_group, destRank);
} 

//typedef void MPI::Comm::Errhandler_fn(MPI::Comm &, int *, ... );
void throwErrHandler(MPI::Comm &comm, int *errCode, ...) {
  int len = 5000;
  char str[len];
  for (int i = 0; i < len; i++) str[i] = 0;
  MPI::Get_error_string(*errCode, str, len);
  std::cerr << "MPI Error: " << str << "\n";
  throw 1;
}

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  //if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeRank << ": " << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  //}
  middleTime = MPI_Wtime();
}

void showDims(const char* name, SubArray<double, 2> *arr) {
  const int N1 = 2;
  for (int d = 0; d < N1; d++) {
    printf("array %s: dim %i [%i..%i] len %i par: [%i..%i] len %i\n", name, d, arr->getDimStart(d), 
         arr->getDimEnd(d), arr->getDimSize(d), arr->parent->getDimStart(d), arr->parent->getDimEnd(d),
         arr->parent->getDimSize(d));
  }
}


//int main(int argc, char** argv) {}
// code generation array demo



// declaring node layouts


typedef struct struct5 {
int v0;
int v1;
bool operator<(const struct5& b) const {
  return (v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1)));
}
bool operator>=(const struct5& b) const {
  return !((v0 < b.v0) || ((v0 == b.v0) && ((v1 < b.v1))));
}
inline bool operator==(const struct5& b) const {
  return (v0 == b.v0) && (v1 == b.v1);
}
inline bool operator!=(const struct5& b) const {
  return !((v0 == b.v0) && (v1 == b.v1));
}
} struct5;
void printVal(const struct5& v) {
printVal("(");
printVal(v.v0);
printVal(", ");
printVal(v.v1);
printVal(")");

}
// HASH NEEDS C++ 11 SO OMITTED FOR NOW
struct eqstruct5 {
inline bool operator()(const struct5& a, const struct5& b) const {
  return (a.v0 == b.v0) && (a.v1 == b.v1);
}
};




CartTopology<2> *top;

void run() {
// No stream called 'main' found in output streams.
  // initializing node layouts


  int v1;
  v1 = N;
  int v39;
  v39 = 0;
  int v40;
  v40 = 1;
  int v146;
  v146 = 3;
  int v147;
  v147 = 0;
  double v148;
  v148 = 0.0;
// genSqMatrix: 
  unsigned int v2[2];
  v2[0] = v1;
  v2[1] = v1;
  
  int v3[2];
  v3[0] = 0;
  v3[1] = 0;
  
  logTime("beforeScatter", &beginTime);

  SubArray<double, 2> v4(MPI::DOUBLE, v2, v3);
  
if (top->isOnRoot()) {
    SubArray<double, 2>::idx_iterator v6(v4.beginIndex());
  SubArray<double, 2>::idx_iterator v7(v4.endIndex());
  
  for (; v6 != v7; v6++) {
    *(v6.getElement()) = 2.0;
  }
}
// scatterArr: 
  int v15[2];
  v15[0] = -1;
  v15[1] = -1;
  
  int v16[2];
  v16[0] = -1;
  v16[1] = 0;
  
  SubArray<double, 2> v14(v4.getElementType());
  
  top->scatterArrayBlocked<double, 2, 1>(&v4, &v14, v15, v16);
// mirrorArr: 
  int v18[2];
  v18[0] = -1;
  v18[1] = -0;
  
  int v19[2];
  v19[0] = -2;
  v19[1] = 0;
  
  SubArray<double, 2> v17(v14.getElementType());
  
static double v20[5000*5000*2];

  top->mirrorArray<double, 2>(&v14, &v17, v18, v19, v20);
// genSqMatrix: 
  unsigned int v21[2];
  v21[0] = v1;
  v21[1] = v1;
  
  int v22[2];
  v22[0] = 0;
  v22[1] = 0;
  
  SubArray<double, 2> v23(MPI::DOUBLE, v21, v22);
  
if (top->isOnRoot()) {
    SubArray<double, 2>::idx_iterator v24(v23.beginIndex());
  SubArray<double, 2>::idx_iterator v25(v23.endIndex());
  
  for (; v24 != v25; v24++) {
    *(v24.getElement()) = 2.0;
  }
}
// scatterArr: 
  int v33[2];
  v33[0] = -1;
  v33[1] = -1;
  
  int v34[2];
  v34[0] = 0;
  v34[1] = -1;
  
  SubArray<double, 2> v32(v23.getElementType());
  
  top->scatterArrayBlocked<double, 2, 1>(&v23, &v32, v33, v34);
// mirrorArr: 
  int v36[2];
  v36[0] = 0;
  v36[1] = -1;
  
  int v37[2];
  v37[0] = 0;
  v37[1] = -2;
  
  SubArray<double, 2> v35(v32.getElementType());
  
static double v38[5000*5000*2];

  top->mirrorArray<double, 2>(&v32, &v35, v36, v37, v38);

    showDims("v32", &v32);
    showDims("v14", &v14);


    showDims("v35", &v35);
    showDims("v17", &v17);


//begin eqJoinArr
  int v46(v35.parent->getDimStart(0));
    int v48(v35.parent->getDimStart(1));
    int v50(v17.parent->getDimStart(0));
    int v52(v17.parent->getDimStart(1));
    int v54(v35.getDimStart(0));
    int v56(v35.getDimStart(1));
    int v58(v17.getDimStart(0));
    int v60(v17.getDimStart(1));
    int v62(v35.getDimEnd(0));
    int v64(v35.getDimEnd(1));
    int v66(v17.getDimEnd(0));
    int v68(v17.getDimEnd(1));
    int v70(v62+1);
    int v72(v64+1);
    int v74(v66+1);
    int v76(v68+1);
    int v78(v35.parent->getDimSize(0));
    int v80(v35.parent->getDimSize(1));
    int v82(v17.parent->getDimSize(0));
    int v84(v17.parent->getDimSize(1));
    int v86(v35.getDimSize(0));
    int v88(v35.getDimSize(1));
    int v90(v17.getDimSize(0));
    int v92(v17.getDimSize(1));
    int v94(1);
    int v96(v94*v80);
    int v98(1);
    int v100(v98*v84);
    int v118(0);
    int v120(0);
    int v122(0);
    int v124(0);
    int v126(v118+v54);
    int v128(v120+v56);
    int v130(v122+v58);
    int v132(v124+v60);
    int v134(v126);
    int v136(v128);
    int v138(v130);
    int v140(v132);
    int v102((v54-v46)*v96);
    int v104((v56-v48)*v94);
    int v106((v58-v50)*v100);
    int v108((v60-v52)*v98);
    int v110(v96*(v70-v46));
    int v112(v94*(v72-v48));
    int v114(v100*(v74-v50));
    int v116(v98*(v76-v52));
    double* v42(v35.parent->elems);
    double* v44(v17.parent->elems);
 

        unsigned int v149[2];
        int v150[2];
        v149[0] = (v62-(v54)+1);
      v149[1] = (v66-(v58)+1);
        v150[0] = (v54);
      v150[1] = (v58);
      static double v151[5000*5000*2];
      int v158(v62-v54+1);
        int v160(v64-v56+1);
        int v162(v68-v60+1);
        int v164(v66-v58+1);
        int v156[2];
        double v169;
      double v170;
    
logTime("beforeloops", &beginTime);
  
#ifdef DEBUG
printf("%i: v118 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "outer1", "0", v118, v102, v110, v96);
printf("%i: v134 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "outer1", "0", v134, v126, v70, v94);
printf("\n");

#endif
#ifdef DEBUG
printf("%i: v122 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "outer2", "0", v122, v106, v114, v100);
printf("%i: v138 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "outer2", "0", v138, v130, v74, v94);
printf("\n");

#endif
#ifdef DEBUG
printf("%i: v120 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "inner", "(1,1)", v120, v104, v112, v94);
printf("%i: v124 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "inner", "(1,1)", v124, v108, v116, v98);
printf("%i: v136 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "inner", "(1,1)", v136, v128, v72, v94);
printf("%i: v140 %s(%s) = %i in [%i..%i:%i]\n", top->getRank(), "inner", "(1,1)", v140, v132, v76, v94);
printf("\n");

#endif
for (v118 = v102, v134 = v126; v118 < v110; v118 += v96, v134 += v94) {
for (v122 = v106, v138 = v130; v122 < v114; v122 += v100, v138 += v94) {
for (v120 = v104, v124 = v108, v136 = v128, v140 = v132; v120 < v112; v120 += v94, v124 += v98, v136 += v94, v140 += v94) {
  double v142(v42[v118 + v120]);
    double v144(v44[v122 + v124]);
    // mapArrStream:
  //decOuts
    double v145;
    //vfCode
    v145 = v142 * v144;
    
  //consume:
      int v168(((v138-v58) + (v164 * (v134-v54))));
        // TupAccNd assignment
        v169 = v145;
        
        v170 = v151[v168];
        v151[v168] = v169 + v170;
      
  }
}
}

logTime("afterloops", &middleTime);



//end eqJoinArr
// print array
// No stream called 'main' found in output streams.

}



int main(int argc, char **argv) {
  for (int i = 0; i < 1000; i++) zeros[i] = 0;
  MPI::Init(argc, argv);
  MPI::Errhandler errHandler = MPI::COMM_WORLD.Create_errhandler(&throwErrHandler);
  MPI::COMM_WORLD.Set_errhandler(errHandler);
  std::cout << "My rank is " << MPI::COMM_WORLD.Get_rank() << "\n";

  // show if debugging
  #ifdef DEBUG
  std::cerr << "DEBUGGING!\n";
  #endif

  // init consts
  int Nproc = MPI::COMM_WORLD.Get_size();

  // the current date/time
  time_t timer = time(NULL);
  struct tm* dateTime = localtime(&timer);
  char *strTime = asctime(dateTime);
  char* strPtr = strTime;
  while (*strPtr != '\0') { if (*strPtr == '\n') *strPtr = '\0'; strPtr++; }
  
  // make topology
  if (argc <= 2) {
    std::cerr << "Must give sizes of x and y topology dims on the command line.\n";
    return -1;
  }
  unsigned int dims[2];
  dims[0] = atoi (argv[1]); 
  dims[1] = atoi (argv[2]);
  top = new CartTopology<2>(dims);

  // make log header
  snprintf(timeHeader, MAX_TIME_HEADER, "%i: %s\t%i\t%i\t%i\t%i", MPI::COMM_WORLD.Get_rank(), strTime, N, dims[0], dims[1], Nproc);

  // time the computation
  initTime();
  try { run(); }
  catch (MPI::Exception e) { std::cerr << "MPI ERROR: " << e.Get_error_code() << " - " << e.Get_error_string() << std::endl; throw 1; }
  logTime("total", &beginTime);

  delete top;
  errHandler.Free();
  MPI::Finalize();
  return 0;
}
